/*
	main.c
	main of wavelet image resize program.
	Copyright(c) & programmed 2004 by NAKATA H. (webmaster@nakata-jp.org)
	$Header: /home/nakata/programs/photo/wavelet1/RCS/main.c,v 1.2 2004/07/11 04:52:53 nakata Exp $

	this program comes with ABSOLUTELY NO WARRANTY.
	This is free software. You can redistribute, customize and sell it.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "wavelet1.h"

static FLOAT_TYPE	*image_red, *image_green, *image_blue;
static int	width, height;

static void	read_bmp( FILE * );
static void	write_bmp( FILE *, int );

extern void		half_size( FLOAT_TYPE *, int, int );
extern FLOAT_TYPE	*double_size( FLOAT_TYPE *, int, int );

/* main section of this program	---------------------------------------------*/
int	main( int argc, char *argv[] )
{
	int	save_width;
	FLOAT_TYPE	*p;
	FILE	*fp;

	if ( argc != 2 && argc != 3 ) {
		fprintf(stderr,
			"usage: %s input_file [output_file]\n", argv[0] );
		exit( 1 );
	}

	if ( (fp = fopen(argv[1], "rb")) == NULL ) {
		fprintf(stderr, "fopen(%s) failed.\n", argv[1] );
		exit( 1 );
	}
	read_bmp( fp );
	fclose( fp );

	if ( strstr( argv[0], "half" ) != NULL ) {
		half_size( image_red, width, height );
		half_size( image_green, width, height );
		half_size( image_blue, width, height );
		save_width = (width + 1) / 2;
		height = (height + 1) / 2;
	} else {
		p = double_size( image_red, width, height );
		free( image_red );
		image_red = p;
		p = double_size( image_green, width, height );
		free( image_green );
		image_green = p;
		p = double_size( image_blue, width, height );
		free( image_blue );
		image_blue = p;
		save_width = (width *= 2);
		height *= 2;
	}

	if ( argc == 2 ) {
		fp = stdout;
	} else {
		if ( (fp = fopen(argv[2], "wb")) == NULL ) {
			fprintf(stderr, "fopen(%s) failed.\n", argv[2] );
			exit( 1 );
		}
	}
	write_bmp( fp, save_width );
	fclose( fp );

	return ( 0 );
}

/* Read the image from a 24-bit BMP file.	-----------------------------*/
static void	read_bmp( FILE *ifp )
{
	char	buff[2];
	short	s;
	long	l, offset;
	FLOAT_TYPE	*p_red, *p_green, *p_blue;
	int	x, y, pudlen;

	if ( fread( buff, 1, 2, ifp ) < 2 ) {
		fprintf(stderr, "read_bmp() fread() failed.\n");
		exit ( 1 );
	}
	if ( strncmp( buff, "BM", 2 ) != 0 ) {
		fprintf( stderr, "this is not BMP file.\n");
		exit ( 1 );
	}

	/* file size */
	if ( fread( &l, sizeof(l), 1, ifp ) < 1 ) {
		fprintf(stderr, "read_bmp() fread() failed.\n");
		exit( 1 );
	}

	/* reserved */
	if ( fread( &l, sizeof(l), 1, ifp ) < 1 ) {
		fprintf(stderr, "read_bmp() fread() failed.\n");
		exit( 1 );
	}

	/* offset */
	if ( fread( &l, sizeof(l), 1, ifp ) < 1 ) {
		fprintf(stderr, "read_bmp() fread() failed.\n");
		exit( 1 );
	}
	offset = itohl(l);

	/* header size */
	if ( fread( &l, sizeof(l), 1, ifp ) < 1 ) {
		fprintf(stderr, "read_bmp() fread() failed.\n");
		exit( 1 );
	}

	/* horizontal dots */
	if ( fread( &l, sizeof(l), 1, ifp ) < 1 ) {
		fprintf(stderr, "read_bmp() fread() failed.\n");
		exit( 1 );
	}
	width = itohl(l);

	/* vertical dots */
	if ( fread( &l, sizeof(l), 1, ifp ) < 1 ) {
		fprintf(stderr, "read_bmp() fread() failed.\n");
		exit( 1 );
	}
	height = itohl(l);

	image_red = malloc( width * height * sizeof(FLOAT_TYPE) );
	image_green = malloc( width * height * sizeof(FLOAT_TYPE) );
	image_blue = malloc( width * height * sizeof(FLOAT_TYPE) );
	if ( image_red == NULL || image_green == NULL || image_blue == NULL ) {
		fprintf( stderr, "allocate image buffer failed.\n" );
		exit( 1 );
	}

	/* plane */
	if ( fread( &s, sizeof(s), 1, ifp ) < 1 ) {
		fprintf(stderr, "read_bmp() fread() failed.\n");
		exit( 1 );
	}
	if ( itohs(s) != 1 ) {
		fprintf(stderr, "this program support only plane=1\n" );
		exit ( 1 );
	}

	/* bits/pixel */
	if ( fread( &s, sizeof(s), 1, ifp ) < 1 ) {
		fprintf(stderr, "read_bmp() fread() failed.\n");
		exit( 1 );
	}
	if ( itohs(s) != 24 ) {
		fprintf(stderr, "this program support only 24bit/plane\n" );
		exit ( 1 );
	}

	if ( fseek( ifp, offset, SEEK_SET ) != 0 ) {
		fprintf(stderr, "read_bmp() seek() error.\n" );
		exit ( 1 );
	}

	pudlen = (width * 3) % 4;
	pudlen = (4 - pudlen) & 3;
	for ( y = 0; y < height; y++ ) {
		p_blue = image_blue + (height - 1 - y) * width;
		p_green = image_green + (height - 1 - y) * width;
		p_red = image_red + (height - 1 - y) * width;
		for ( x = 0; x < width; x++ ) {
			*p_blue++ = fgetc( ifp );
			*p_green++ = fgetc( ifp );
			*p_red++ = fgetc( ifp );
		}
		/* adjust to word align */
		for ( x = 0; x < pudlen; x++ ) {
			fgetc( ifp );
		}
	}
}

/* Write the image to a 24-bit BMP file.	-----------------------------*/
static void	write_bmp( FILE *ofp, int save_width )
{
	int	c;
	short	s;
	long	l;
	FLOAT_TYPE	*p_red, *p_green, *p_blue;
	int	x, y, pudlen;

	if ( fwrite( "BM", 1, 2, ofp ) < 2 ) {
		fprintf(stderr, "write_bmp() fwrite() failed.\n");
		exit ( 1 );
	}

	pudlen = (save_width * 3) % 4;
	pudlen = (4 - pudlen) & 3;
	/* file size */
	l = (save_width * 3 + pudlen) * height + 0x36;
	l = htoil(l);
	if ( fwrite( &l, sizeof(l), 1, ofp ) < 1 ) {
		fprintf(stderr, "write_bmp() fwrite() failed.\n");
		exit( 1 );
	}

	/* reserved */
	l = 0;
	if ( fwrite( &l, sizeof(l), 1, ofp ) < 1 ) {
		fprintf(stderr, "write_bmp() fwrite() failed.\n");
		exit( 1 );
	}

	/* offset */
	l = 0x36;
	l = htoil(l);
	if ( fwrite( &l, sizeof(l), 1, ofp ) < 1 ) {
		fprintf(stderr, "write_bmp() fwrite() failed.\n");
		exit( 1 );
	}

	/* header size */
	l = 0x28;
	l = htoil(l);
	if ( fwrite( &l, sizeof(l), 1, ofp ) < 1 ) {
		fprintf(stderr, "write_bmp() fwrite() failed.\n");
		exit( 1 );
	}

	/* horizontal dots */
	l = htoil(save_width);
	if ( fwrite( &l, sizeof(l), 1, ofp ) < 1 ) {
		fprintf(stderr, "write_bmp() fwrite() failed.\n");
		exit( 1 );
	}

	/* vertical dots */
	l = htoil(height);
	if ( fwrite( &l, sizeof(l), 1, ofp ) < 1 ) {
		fprintf(stderr, "write_bmp() fwrite() failed.\n");
		exit( 1 );
	}

	/* plane */
	s = htois(1);
	if ( fwrite( &s, sizeof(s), 1, ofp ) < 1 ) {
		fprintf(stderr, "write_bmp() fwrite() failed.\n");
		exit( 1 );
	}

	/* bits/pixel */
	s = htois(24);
	if ( fwrite( &s, sizeof(s), 1, ofp ) < 1 ) {
		fprintf(stderr, "write_bmp() fwrite() failed.\n");
		exit( 1 );
	}

	/* compression */
	l = 0;
	if ( fwrite( &l, sizeof(l), 1, ofp ) < 1 ) {
		fprintf(stderr, "write_bmp() fwrite() failed.\n");
		exit( 1 );
	}

	/* bitmap data */
	l = 0;
	if ( fwrite( &l, sizeof(l), 1, ofp ) < 1 ) {
		fprintf(stderr, "write_bmp() fwrite() failed.\n");
		exit( 1 );
	}

	/* horizontal resolution*/
	l = htoil(0x6893);
	if ( fwrite( &l, sizeof(l), 1, ofp ) < 1 ) {
		fprintf(stderr, "write_bmp() fwrite() failed.\n");
		exit( 1 );
	}

	/* vertical resolution*/
	l = htoil(0x6893);
	if ( fwrite( &l, sizeof(l), 1, ofp ) < 1 ) {
		fprintf(stderr, "write_bmp() fwrite() failed.\n");
		exit( 1 );
	}

	/* colors */
	l = 0;
	if ( fwrite( &l, sizeof(l), 1, ofp ) < 1 ) {
		fprintf(stderr, "write_bmp() fwrite() failed.\n");
		exit( 1 );
	}

	/* important colors */
	l = 0;
	if ( fwrite( &l, sizeof(l), 1, ofp ) < 1 ) {
		fprintf(stderr, "write_bmp() fwrite() failed.\n");
		exit( 1 );
	}

	for ( y = 0; y < height; y++ ) {
		p_blue = image_blue + (height - 1 - y) * width;
		p_green = image_green + (height - 1 - y) * width;
		p_red = image_red + (height - 1 - y) * width;
		for ( x = 0; x < save_width; x++ ) {
			c = *p_blue++;
			if ( c < 0 ) {
				c = 0;
			} else if ( c > MAX_UCHAR ) {
				c = MAX_UCHAR;
			}
			fputc( c, ofp );
			c = *p_green++;
			if ( c < 0 ) {
				c = 0;
			} else if ( c > MAX_UCHAR ) {
				c = MAX_UCHAR;
			}
			fputc( c, ofp );
			c = *p_red++;
			if ( c < 0 ) {
				c = 0;
			} else if ( c > MAX_UCHAR ) {
				c = MAX_UCHAR;
			}
			fputc( c, ofp );
		}
		/* adjust to word align */
		for ( x = 0; x < pudlen; x++ ) {
			fputc( 0, ofp );
		}
	}
}
/* End Of File	**************************************************************/
