/*
	wavelet.c
	wavelet section of wavelet image resize program.
	Copyright(c) & programmed 2004 by NAKATA H. (webmaster@nakata-jp.org)
	$Header: /home/nakata/programs/photo/wavelet1/RCS/wavelet.c,v 1.2 2004/07/11 04:52:53 nakata Exp $

	this program comes with ABSOLUTELY NO WARRANTY.
	This is free software. You can redistribute, customize and sell it.
 */

#include <stdio.h>
#include <stdlib.h>
#include "wavelet1.h"

static void	wavelet( FLOAT_TYPE *, int, int );
static void	rwavelet( FLOAT_TYPE *, int, int );

static FLOAT_TYPE	*buff = NULL;
static size_t	buff_size = 0;

void	half_size( FLOAT_TYPE *p_image, int width, int height )
{
	int	i;

	/* vertical wavelet */
	for ( i = 0; i < width; i++ ) {
		wavelet( p_image + i, height, width );
	}
	/* horizontal wavelet */
	for ( i = 0; i < (height + 1) / 2; i++ ) {
		wavelet( p_image + i * width, width, 1 );
	}
}

FLOAT_TYPE	*double_size( FLOAT_TYPE *p_image, int width, int height )
{
	int		x, y;
	FLOAT_TYPE	*p, *p2, *new_buff;

	if ( (new_buff = (FLOAT_TYPE *)malloc( sizeof(FLOAT_TYPE) * width * 2
					       * height * 2 )) == NULL ) {
		fprintf(stderr, "malloc() failed.\n" );
		exit( 1 );
	}

	p = new_buff;
	p2 = p_image;
	for ( y = 0; y < height; y++ ) {
		for ( x = 0; x < width; x++ ) {
			*p++ = *p2++;
		}
		p += width;
	}
	/* horizontal reverse wavelet */
	for ( y = 0; y < height; y++ ) {
		rwavelet( new_buff + y * width * 2, width, 1 );
	}
	/* vertical reverse wavelet */
	for ( x = 0; x < width * 2; x++ ) {
		rwavelet( new_buff + x, height, width * 2 );
	}
	return (new_buff);
}

#define	BEFORE_SHIFT	4	/* start index i0 == 0 (even) */
#define	ALPHA		((FLOAT_TYPE)-1.586134342059924)
#define	BETA		((FLOAT_TYPE)-0.052980118572961)
#define	GAMMA		((FLOAT_TYPE)0.882911075530934)
#define	DELTA		((FLOAT_TYPE)0.443506852043971)
#define	K		((FLOAT_TYPE)1.230174104914001)

/* wavelet function	-----------------------------------------------------*/
static void	wavelet( FLOAT_TYPE *source, int length, int skip )
{
	int		after_shift;
	int		i;
	size_t		size;
	FLOAT_TYPE	*p;

	after_shift = (length % 2) == 0 ? 3 : 4;
	size = sizeof(FLOAT_TYPE) * (BEFORE_SHIFT + length + after_shift);
	if ( buff == NULL ) {
		buff = (FLOAT_TYPE *)malloc(size);
		buff_size = size;
	} else if ( buff_size < size ) {
		buff = (FLOAT_TYPE *)realloc(buff, size);
		buff_size = size;
	}
	if ( buff == NULL ) {
		fprintf(stderr, "malloc() failed.\n" );
		exit( 1 );
	}
	p = source;
	for ( i = 0; i < length; i++ ) {
		buff[BEFORE_SHIFT + i] = *p;
		p += skip;
	}

	/* periodic symmetric extension */
	for ( i = 0; i < BEFORE_SHIFT; i++ ) {
		buff[i] = buff[BEFORE_SHIFT * 2 - i];
	}
	for ( i = 0; i < after_shift; i++ ) {
		buff[BEFORE_SHIFT + length + i]
			= buff[BEFORE_SHIFT + length - 2 - i];
	}

	for ( i = 0; i < length / 2 + 3; i++ ) {
		buff[2 * i + 1] += ALPHA * ( buff[2 * i] + buff[2 * i + 2] );
	}
	for ( i = 0; i < length / 2 + 3; i++ ) {
		buff[2 * i] += BETA * ( buff[2 * i - 1] + buff[2 * i + 1] );
	}
	for ( i = 1; i < length / 2 + 2; i++ ) {
		buff[2 * i + 1] += GAMMA * ( buff[2 * i] + buff[2 * i + 2] );
	}
	for ( i = 2; i < length / 2 + 2; i++ ) {
		buff[2 * i] += DELTA * ( buff[2 * i - 1] + buff[2 * i + 1] );
		buff[2 * i] /= K;
	}

	p = source;
	for ( i = 0; i < (length + 1)/2; i++ ) {
		*p = buff[BEFORE_SHIFT + 2 * i];
		p += skip;
	}
}
#undef	BEFORE_SHIFT

#define	BEFORE_SHIFT	3
#define	AFTER_SHIFT	3
/* reverse wavelet function	---------------------------------------------*/
static void	rwavelet( FLOAT_TYPE *source, int length, int skip )
{
	int		i;
	size_t		size;
	FLOAT_TYPE	*p;

	size = sizeof(FLOAT_TYPE) * (BEFORE_SHIFT + length * 2 + AFTER_SHIFT);
	if ( buff == NULL ) {
		buff = (FLOAT_TYPE *)malloc(size);
		buff_size = size;
	} else if ( buff_size < size ) {
		buff = (FLOAT_TYPE *)realloc(buff, size);
		buff_size = size;
	}
	if ( buff == NULL ) {
		fprintf(stderr, "malloc() failed.\n" );
		exit( 1 );
	}

	p = source;
	for ( i = 0; i < length; i++ ) {
		buff[BEFORE_SHIFT + 2 * i] = K * *p;
		buff[BEFORE_SHIFT + 2 * i + 1] = 0.0;
		p += skip;
	}

	/* periodic symmetric extension */
	for ( i = 0; i < BEFORE_SHIFT; i++ ) {
		buff[i] = buff[BEFORE_SHIFT * 2 - i];
	}
	for ( i = 0; i < AFTER_SHIFT; i++ ) {
		buff[BEFORE_SHIFT + length * 2 + i]
			= buff[BEFORE_SHIFT + length * 2 - 2 - i];
	}

	for ( i = 0; i < length + 3; i++ ) {
		buff[2 * i + 1] -= DELTA * ( buff[2 * i] + buff[2 * i + 2] );
	}
	for ( i = 0; i < length + 2; i++ ) {
		buff[2 * i + 2] -= GAMMA * ( buff[2 * i + 1]
					     + buff[2 * i + 3] );
	}
	for ( i = 1; i < length + 2; i++ ) {
		buff[2 * i + 1] -= BETA * ( buff[2 * i] + buff[2 * i + 2] );
	}
	for ( i = 1; i < length + 1; i++ ) {
		buff[2 * i + 2] -= ALPHA * ( buff[2 * i + 1]
					     + buff[2 * i + 3] );

	}

	p = source;
	for ( i = 0; i < length * 2; i++ ) {
		*p = buff[BEFORE_SHIFT + i];
		p += skip;
	}
}
/* End Of File	**************************************************************/
